from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

import yaml
from ceci.config import StageParameter
from rail.core.configurable import Configurable
from rail.core.stage import RailPipeline
from rail.utils import catalog_utils

if TYPE_CHECKING:
    from .project import RailProject


def inform_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,  # pylint: disable=unused-argument
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the inform pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    flavor = kwargs.pop("flavor", "baseline")
    for key, val in input_file_tags.items():
        input_file_flavor = val.get("flavor", flavor)
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor, val["tag"], **kwargs
        )
    return input_files


def inform_sompz_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,  # pylint: disable=unused-argument
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the inform pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    flavor = kwargs.pop("flavor", "baseline")
    for key, val in input_file_tags.items():
        input_file_flavor = val.get("flavor", flavor)
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor, val["tag"], **kwargs
        )
    return input_files


def inform_recalib_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,  # pylint: disable=unused-argument
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the inform pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    flavor = kwargs.pop("flavor", "baseline")
    for key, val in input_file_tags.items():
        input_file_flavor = val.get("flavor", flavor)
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor, val["tag"], **kwargs
        )
    pdfs_dir = sink_dir
    pz_algorithms = project.get_pzalgorithms()
    for pz_algo_ in pz_algorithms.keys():
        input_files[f"input_{pz_algo_}"] = os.path.join(
            pdfs_dir, f"output_estimate_{pz_algo_}.hdf5"
        )
    return input_files


def inform_somlike_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,  # pylint: disable=unused-argument
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the inform pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    flavor = kwargs.pop("flavor", "baseline")
    for key, val in input_file_tags.items():
        input_file_flavor = val.get("flavor", flavor)
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor, val["tag"], **kwargs
        )
    return input_files


def estimate_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the estimate pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    kwcopy = kwargs.copy()
    flavor = kwcopy.pop("flavor", "baseline")
    local_input_tag = kwcopy.pop("input_tag", None)
    if local_input_tag:
        input_files["sink_dir"] = os.path.join(sink_dir, local_input_tag)
    else:  # pragma: no cover
        input_files["sink_dir"] = sink_dir
    for key, val in input_file_tags.items():
        input_file_flavor = kwargs.get("flavor", val.get("flavor", flavor))
        input_tag = kwargs.get("input_tag", val["tag"])
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor,
            input_tag,
            **kwcopy,
        )

    pz_algorithms = project.get_pzalgorithms()
    for pz_algo_ in pz_algorithms.keys():
        input_files[f"model_{pz_algo_}"] = os.path.join(
            project.get_path("ceci_output_dir", flavor=input_file_flavor, **kwcopy),
            f"model_inform_{pz_algo_}.pkl",
        )
    return input_files


def estimate_sompz_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the estimate pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    kwcopy = kwargs.copy()
    flavor = kwcopy.pop("flavor", "baseline")
    local_input_tag = kwcopy.pop("input_tag", None)
    if local_input_tag:
        input_files["sink_dir"] = os.path.join(sink_dir, local_input_tag)
    else:  # pragma: no cover
        input_files["sink_dir"] = sink_dir
    for key, val in input_file_tags.items():
        input_file_flavor = kwargs.get("flavor", val.get("flavor", flavor))
        input_tag = kwargs.get("input_tag", val["tag"])
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor,
            input_tag,
            **kwcopy,
        )

    for field_ in ["wide", "deep"]:
        input_files[f"{field_}_model"] = os.path.join(
            project.get_path("ceci_output_dir", flavor=input_file_flavor, **kwcopy),
            f"model_som_informer_{field_}.pkl",
        )
    return input_files


def estimate_recalib_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the estimate pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    kwcopy = kwargs.copy()
    flavor = kwcopy.pop("flavor", "baseline")
    local_input_tag = kwcopy.pop("input_tag", None)
    if local_input_tag:
        input_files["sink_dir"] = os.path.join(sink_dir, local_input_tag)
    else:  # pragma: no cover
        input_files["sink_dir"] = sink_dir
    for key, val in input_file_tags.items():
        input_file_flavor = kwargs.get("flavor", val.get("flavor", flavor))
        input_tag = kwargs.get("input_tag", val["tag"])
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor,
            input_tag,
            **kwcopy,
        )

    pz_algorithms = project.get_pzalgorithms()
    ceci_dir = project.get_path("ceci_output_dir", flavor=input_file_flavor, **kwcopy)
    for pz_algo_ in pz_algorithms.keys():
        input_files[f"input_{pz_algo_}"] = os.path.join(
            ceci_dir, f"output_estimate_{pz_algo_}.hdf5"
        )
        for recalib_algo_ in ["pz_max_cell_p", "pz_mode"]:
            input_files[f"model_{pz_algo_}_{recalib_algo_}"] = os.path.join(
                ceci_dir,
                f"model_inform_{pz_algo_}_{recalib_algo_}.pkl",
            )

    return input_files


def somlike_recalib_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the somlike recalib pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    kwcopy = kwargs.copy()
    flavor = kwcopy.pop("flavor", "baseline")

    models_dir = sink_dir
    pz_algorithms = project.get_pzalgorithms()
    for pz_algo_ in pz_algorithms.keys():
        for field_ in ["deep", "wide"]:
            model_file = f"model_pz_informer_{pz_algo_}_{field_}.pkl"
            input_files[f"model_{pz_algo_}_{field_}"] = os.path.join(
                models_dir, model_file
            )

    local_input_tag = kwcopy.pop("input_tag", None)
    if local_input_tag:  # pragma: no cover
        input_files["sink_dir"] = os.path.join(sink_dir, local_input_tag)
    else:
        input_files["sink_dir"] = sink_dir
    for key, val in input_file_tags.items():
        input_file_flavor = kwargs.get("flavor", val.get("flavor", flavor))
        input_tag = kwargs.get("input_tag", val["tag"])
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor,
            input_tag,
            **kwcopy,
        )

    return input_files


def evaluate_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the evalute pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    flavor = kwargs.pop("flavor", "baseline")
    for key, val in input_file_tags.items():
        input_file_flavor = val.get("flavor", flavor)
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor, val["tag"], **kwargs
        )

    pdfs_dir = sink_dir
    pz_algorithms = project.get_pzalgorithms()
    for pz_algo_ in pz_algorithms.keys():
        input_files[f"input_evaluate_{pz_algo_}"] = os.path.join(
            pdfs_dir, f"estimate_output_{pz_algo_}.hdf5"
        )
    return input_files


def pz_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,  # pylint: disable=unused-argument
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the pz pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    flavor = kwargs.pop("flavor")
    for key, val in input_file_tags.items():
        input_file_flavor = val.get("flavor", flavor)
        if input_file_flavor != flavor:  # pragma: no cover
            input_file_flavor = flavor
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor, val["tag"], **kwargs
        )
    return input_files


def tomography_input_callback(
    project: RailProject,
    pipeline_name: str,
    sink_dir: str,
    **kwargs: Any,
) -> dict[str, str]:
    """Make dict of input tags and paths for the tomography pipeline

    Parameters
    ----------
    project: RailProject
        Object with project configuration

    pipeline_name: str
        Name of the pipeline to run

    sink_dir: str
        Path to output directory

    kwargs: Any
        Additional parameters to specify pipeline, e.g., flavor, selection, ...

    Returns
    -------
    dict[str, str]:
        Dictionary of input file tags and paths
    """
    pipeline_info = project.get_pipeline(pipeline_name)
    input_files = {}
    input_file_tags = pipeline_info["input_file_templates"]
    flavor = kwargs.pop("flavor")
    selection = kwargs.get("selection")
    for key, val in input_file_tags.items():
        input_file_flavor = val.get("flavor", flavor)
        input_files[key] = project.get_file_for_flavor(
            input_file_flavor, val["tag"], selection=selection
        )

    pdfs_dir = sink_dir
    pz_algorithms = project.get_pzalgorithms()
    for pz_algo_ in pz_algorithms.keys():
        input_files[f"input_{pz_algo_}"] = os.path.join(
            pdfs_dir, f"output_estimate_{pz_algo_}.hdf5"
        )

    return input_files


def truth_to_observed_convert_commands(
    sink_dir: str, **_kwargs: Any
) -> list[list[str]]:
    convert_command = [
        "tables-io",
        "convert",
        "--input",
        f"{sink_dir}/output_dereddener_errors.pq",
        "--output",
        f"{sink_dir}/output.hdf5",
    ]
    convert_commands = [convert_command]
    return convert_commands


def prepare_convert_commands(sink_dir: str, **_kwargs: Any) -> list[list[str]]:
    convert_command = [
        "tables-io",
        "convert",
        "--input",
        f"{sink_dir}/output_deredden.pq",
        "--output",
        f"{sink_dir}/output.hdf5",
    ]
    convert_commands = [convert_command]
    return convert_commands


def photometric_errors_convert_commands(
    sink_dir: str, **_kwargs: Any
) -> list[list[str]]:
    convert_command = [
        "tables-io",
        "convert",
        "--input",
        f"{sink_dir}/output_dereddener_errors.pq",
        "--output",
        f"{sink_dir}/output.hdf5",
    ]
    convert_commands = [convert_command]
    return convert_commands


def spectroscopic_selection_convert_commands(
    sink_dir: str, **kwargs: Any
) -> list[list[str]]:
    convert_commands = []
    spec_selections = kwargs.get("spec_selections")
    assert isinstance(spec_selections, list)
    for spec_selection_ in spec_selections:
        convert_command = [
            "tables-io",
            "convert",
            "--input",
            f"{sink_dir}/output_select_{spec_selection_}.pq",
            "--output",
            f"{sink_dir}/output_select_{spec_selection_}.hdf5",
        ]
        convert_commands.append(convert_command)
    return convert_commands


def blending_convert_commands(sink_dir: str, **_kwargs: Any) -> list[list[str]]:
    convert_command = [
        "tables-io",
        "convert",
        "--input",
        f"{sink_dir}/output_blended.pq",
        "--output",
        f"{sink_dir}/output_blended.hdf5",
    ]
    convert_commands = [convert_command]
    return convert_commands


INPUT_CALLBACK_DICT = dict(
    inform=inform_input_callback,
    estimate=estimate_input_callback,
    evaluate=evaluate_input_callback,
    inform_sompz=inform_sompz_input_callback,
    estimate_sompz=estimate_sompz_input_callback,
    inform_recalib=inform_recalib_input_callback,
    estimate_recalib=estimate_recalib_input_callback,
    inform_somlike=inform_somlike_input_callback,
    somlike_recalib=somlike_recalib_input_callback,
    pz=pz_input_callback,
    tomography=tomography_input_callback,
)


CATALOG_CONVERT_COMMANDS_DICT = dict(
    truth_to_observed=truth_to_observed_convert_commands,
    prepare=prepare_convert_commands,
    photometric_errors=photometric_errors_convert_commands,
    spec_selection=spectroscopic_selection_convert_commands,
    blending=blending_convert_commands,
)


class RailPipelineTemplate(Configurable):
    """Simple class for holding a pipeline configuraiton"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Pipeline name"),
        pipeline_class=StageParameter(
            str,
            None,
            fmt="%s",
            required=True,
            msg="Full class name for Pipeline",
        ),
        input_catalog_template=StageParameter(
            str,
            "",
            fmt="%s",
            msg="Template to use for input catalog",
        ),
        output_catalog_template=StageParameter(
            str,
            "",
            fmt="%s",
            msg="Template to use for output catalog",
        ),
        input_catalog_basename=StageParameter(
            str,
            "",
            fmt="%s",
            msg="Basename to use for input catalog",
        ),
        input_file_templates=StageParameter(
            dict,
            {},
            fmt="%s",
            msg="Templates to use for input files",
        ),
        kwargs=StageParameter(
            dict,
            {},
            fmt="%s",
            msg="Keywords to provide Pipeline c'tor",
        ),
    )

    yaml_tag = "PipelineTemplate"

    def __init__(self, **kwargs: Any) -> None:
        """C'tor

        Parameters
        ----------
        **kwargs: Any
            Configuration parameters for this RailPipelineTemplate, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)

    def __repr__(self) -> str:
        return f"{self.config.pipeline_class}"

    def make_instance(
        self,
        project: RailProject,
        flavor: str,
        pipeline_overrides: dict[str, Any],
    ) -> RailPipelineInstance:
        kwargs_copy = self.config.kwargs.copy()
        if "kwargs" in pipeline_overrides:  # pragma: no cover
            kwargs_update = pipeline_overrides.pop("kwargs", {})
            kwargs_copy.update(**kwargs_update)

        path = project.get_path(
            "pipeline_path", pipeline=self.config.name, flavor=flavor
        )
        return RailPipelineInstance(
            name=f"{self.config.name}_{flavor}",
            path=path,
            pipeline_template=self.config.name,
            flavor=flavor,
            kwargs=kwargs_copy,
            pipeline_overrides=pipeline_overrides,
        )


class RailPipelineInstance(Configurable):
    """Simple class for holding a pipeline configuraiton"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Pipeline name"),
        path=StageParameter(str, None, fmt="%s", msg="Path to pipeline file"),
        pipeline_template=StageParameter(
            str,
            None,
            fmt="%s",
            required=True,
            msg="Name of PipelineTemplate to use",
        ),
        flavor=StageParameter(str, None, fmt="%s", msg="Pipeline flavor"),
        kwargs=StageParameter(
            dict,
            {},
            fmt="%s",
            msg="Keywords to provide Pipeline c'tor",
        ),
        pipeline_overrides=StageParameter(
            dict,
            {},
            fmt="%s",
            msg="Parameters to override from template",
        ),
    )

    yaml_tag = "PipelineInstance"

    def __init__(self, **kwargs: Any):
        """C'tor

        Parameters
        ----------
        kwargs: Any
            Configuration parameters for this RailPipelineInstance, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)

    def __repr__(self) -> str:
        return f"{self.config.pipeline_template} {self.config.path}"

    def build(
        self,
        project: RailProject,
    ) -> int:
        output_dir = project.get_common_path("project_scratch_dir")
        pipeline_template = project.get_pipeline(self.config.pipeline_template)
        pipeline_class = pipeline_template.config.pipeline_class

        pipe_out_dir = os.path.dirname(self.config.path)
        try:
            os.makedirs(pipe_out_dir)
        except FileExistsError:
            pass

        pipeline_kwargs = self.config.kwargs.copy()

        if self.config.pipeline_overrides:  # pragma: no cover
            copy_overrides = self.config.pipeline_overrides.copy()

            stages_config = os.path.join(
                pipe_out_dir, f"{self.config.name}_overrides.yml"
            )

            kwarg_overrides = copy_overrides.pop("kwargs", {})
            pipeline_kwargs.update(**kwarg_overrides)

            with open(stages_config, "w", encoding="utf-8") as fout:
                yaml.dump(copy_overrides, fout)
        else:
            stages_config = None

        for key, val in pipeline_kwargs.items():
            if key == "selectors":
                temp_dict = project.get_spec_selections()
            elif key == "algorithms":
                temp_dict = project.get_pzalgorithms()
            elif key == "classifiers":
                temp_dict = project.get_classifiers()
            elif key == "summarizers":
                temp_dict = project.get_summarizers()
            elif key == "error_models":
                temp_dict = project.get_error_models()
            else:
                continue
            if "all" in val:
                pipeline_kwargs[key] = temp_dict
            else:
                pipeline_kwargs[key] = {
                    algo_name_: temp_dict[algo_name_] for algo_name_ in val
                }

        catalog_tag = project.get_flavor(self.config.flavor).get("catalog_tag", None)
        if catalog_tag:
            try:
                catalog_utils.apply_defaults(catalog_tag)
            except KeyError:  # pragma: no cover
                tokens = catalog_tag.split(".")
                module_name = ".".join(tokens[:-1])
                class_name = tokens[-1]
                __import__(module_name)
                catalog_utils.CatalogConfigBase.apply_class(class_name)

        tokens = pipeline_class.split(".")
        module = ".".join(tokens[:-1])
        class_name = tokens[-1]
        log_dir = f"{pipe_out_dir}/logs/{self.config.name}"

        print(f"Writing {self.config.path}")

        __import__(module)
        RailPipeline.build_and_write(
            class_name,
            self.config.path,
            None,
            stages_config,
            output_dir,
            log_dir,
            **pipeline_kwargs,
        )

        return 0

    def get_input_files(
        self,
        project: RailProject,
        **kwargs: Any,
    ) -> dict[str, str]:
        """Get the input files needed to run this instandce

        Parameters
        ----------
        project: RailProject
            Object with project configuration

        **kwargs:
            Additional parameters to specify pipeline, e.g., flavor, selection, ...

        Returns
        -------
        dict[str, str]
            Input files, keyed by label
        """

        pipeline_name = self.config.pipeline_template
        sink_dir = project.get_path(
            "ceci_output_dir", flavor=self.config.flavor, **kwargs
        )
        input_callback = INPUT_CALLBACK_DICT[pipeline_name]
        input_files = input_callback(
            project, pipeline_name, sink_dir, flavor=self.config.flavor, **kwargs
        )
        input_files.setdefault("sink_dir", sink_dir)
        return input_files

    def make_pipeline_single_input_command(
        self,
        project: RailProject,
        **kwargs: Any,
    ) -> list[str]:
        """Make the command to run a single instance

        Parameters
        ----------
        project: RailProject
            Object with project configuration

        **kwargs:
            Additional parameters to specify pipeline, e.g., flavor, selection, ...

        Returns
        -------
        list[str]
            Commands to run
        """
        pipeline_path = self.config.path
        pipeline_config = pipeline_path.replace(".yaml", "_config.yml")
        input_files = self.get_input_files(project, **kwargs)
        sink_dir = input_files.pop("sink_dir")
        command_line = project.generate_ceci_command(
            pipeline_path=pipeline_path,
            config=pipeline_config,
            inputs=input_files,
            output_dir=sink_dir,
            log_dir=f"{sink_dir}/logs",
        )
        return command_line

    def make_pipeline_catalog_commands(
        self,
        project: RailProject,
        **kwargs: Any,
    ) -> list[tuple[list[list[str]], str]]:
        """Make the command to run a pipeline on an entire catalog

        Parameters
        ----------
        project: RailProject
            Object with project configuration

        kwargs: Any
            Additional parameters to specify pipeline, e.g., flavor, selection, ...

        Returns
        -------
        commands: list[tuple[list[list[str]], str]]
            Commands to run
        """
        pipeline_name = self.config.pipeline_template
        pipeline_info = project.get_pipeline(pipeline_name)
        flavor = self.config.flavor
        pipeline_path = project.get_path(
            "pipeline_path", pipeline=pipeline_name, flavor=flavor, **kwargs
        )

        catalog_convert_commands_function = CATALOG_CONVERT_COMMANDS_DICT[pipeline_name]

        source_catalog_files = project.get_catalog_files(
            pipeline_info.config.input_catalog_template,
            basename=pipeline_info.config.input_catalog_basename,
            flavor=self.config.flavor,
            **kwargs,
        )
        sink_catalog_files = project.get_catalog_files(
            pipeline_info.config.output_catalog_template,
            basename="output.hdf5",
            flavor=self.config.flavor,
            **kwargs,
        )

        all_commands: list[tuple[list[list[str]], str]] = []

        selection = kwargs["selection"]

        for source_catalog, sink_catalog in zip(
            source_catalog_files, sink_catalog_files
        ):
            sink_dir = os.path.dirname(sink_catalog)
            script_path = os.path.join(
                sink_dir,
                f"submit_{pipeline_name}_{selection}_{flavor}.sh",
            )
            ceci_commands = project.generate_ceci_command(
                pipeline_path=pipeline_path,
                config=pipeline_path.replace(".yaml", "_config.yml"),
                inputs=dict(input=source_catalog),
                output_dir=sink_dir,
                log_dir=sink_dir,
            )
            convert_commands = catalog_convert_commands_function(
                sink_dir,
                **kwargs,
            )
            iter_commands = [
                ["mkdir", "-p", f"{sink_dir}"],
                ceci_commands,
                *convert_commands,
            ]
            all_commands.append((iter_commands, script_path))

        return all_commands



from __future__ import annotations

from typing import Any

from ceci.config import StageParameter

from rail.projects import RailProject, path_funcs

from .data_extraction_funcs import (
    get_multi_pz_point_estimate_data,
    get_pz_point_estimate_data,
)
from .dataset import RailDataset
from .dataset_factory import RailDatasetFactory
from .dataset_holder import (
    DatasetSplitMode,
    RailDatasetHolder,
    RailDatasetListHolder,
    RailProjectHolder,
)
from .pz_plotters import RailPZMultiPointEstimateDataset, RailPZPointEstimateDataset


class RailPZPointEstimateDataHolder(RailDatasetHolder):
    """Class to extract true redshifts and one p(z) point estimate
    from a RailProject.

    This will return a dict:

    truth: np.ndarray
        True redshifts

    pointEstimate: np.ndarray
        Point estimates of the true redshifts
    """

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Dataset name"),
        project=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject name"
        ),
        selection=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject data selection"
        ),
        flavor=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject analysis flavor"
        ),
        tag=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject file tag"
        ),
        algo=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject algorithm"
        ),
    )

    extractor_inputs: dict = {
        "project": RailProject,
        "selection": str,
        "flavor": str,
        "tag": str,
        "algo": str,
    }

    output_type: type[RailDataset] = RailPZPointEstimateDataset

    def __init__(self, **kwargs: Any):
        RailDatasetHolder.__init__(self, **kwargs)
        self._project: RailProject | None = None

    def __repr__(self) -> str:
        ret_str = (
            f"{self.__class__.__name__} "
            "( "
            f"{self.config.project}, "
            f"{self.config.selection}_{self.config.flavor}_{self.config.tag}_{self.config.algo}"
            ")"
        )
        return ret_str

    def _get_data(self, **kwargs: Any) -> dict[str, Any] | None:
        return get_pz_point_estimate_data(**kwargs)

    def get_extractor_inputs(self) -> dict[str, Any]:
        if self._project is None:
            self._project = RailDatasetFactory.get_project(
                self.config.project
            ).resolve()
        the_extractor_inputs = dict(
            project=self._project,
            selection=self.config.selection,
            flavor=self.config.flavor,
            tag=self.config.tag,
            algo=self.config.algo,
        )
        self._validate_extractor_inputs(**the_extractor_inputs)
        return the_extractor_inputs

    @classmethod
    def generate_dataset_dict(
        cls,
        **kwargs: Any,
    ) -> tuple[
        list[RailProjectHolder], list[RailDatasetHolder], list[RailDatasetListHolder]
    ]:
        """
        Parameters
        ----------
        **kwargs
            Set Notes

        Notes
        -----
        dataset_list_name: str
            Name for the resulting DatasetList

        project_file: str
            Config file for project to inspect

        selections: list[str]
            Selections to use

        flavors: list[str]
            Flavors to use

        Returns
        -------
        list[RailProjectHolder]
            Underlying RailProjects

        list[RailDatasetHolder]
            Extracted datasets

        list[RailDatasetListHolder]
            Extracted dataset lists
        """
        dataset_list_name: str | None = kwargs.get("dataset_list_name")
        project_file = kwargs["project_file"]
        project = RailProject.load_config(project_file)
        selections = kwargs.get("selections")
        flavors = kwargs.get("flavors")
        split_mode = kwargs.get("split_mode", DatasetSplitMode.by_algo)

        flavor_dict = project.get_flavors()
        if flavors is None or "all" in flavors:
            flavors = list(flavor_dict.keys())
        if selections is None or "all" in selections:
            selections = list(project.get_selections().keys())

        project_name = project.name
        if not dataset_list_name:
            dataset_list_name = f"{project_name}_pz"

        projects: list[RailProjectHolder] = []
        datasets: list[RailDatasetHolder] = []
        dataset_lists: list[RailDatasetListHolder] = []

        projects.append(
            RailProjectHolder(
                name=project_name,
                yaml_file=project_file,
            )
        )

        dataset_list_dict: dict[str, list[str]] = {}
        dataset_key = dataset_list_name
        if split_mode == DatasetSplitMode.no_split:
            dataset_list_dict[dataset_key] = []

        for key in flavors:
            val = flavor_dict[key]
            pipelines = val["pipelines"]
            if "all" not in pipelines and "pz" not in pipelines:  # pragma: no cover
                continue
            try:
                algos = val["pipeline_overrides"]["default"]["kwargs"]["algorithms"]
            except KeyError:
                algos = list(project.get_pzalgorithms().keys())

            for selection_ in selections:
                if split_mode == DatasetSplitMode.by_flavor:
                    dataset_key = f"{dataset_list_name}_{selection_}_{key}"
                    dataset_list_dict[dataset_key] = []

                for algo_ in algos:
                    if split_mode == DatasetSplitMode.by_algo:
                        dataset_key = f"{dataset_list_name}_{selection_}_{algo_}"
                        if dataset_key not in dataset_list_dict:
                            dataset_list_dict[dataset_key] = []

                    path = path_funcs.get_ceci_pz_output_path(
                        project,
                        selection=selection_,
                        flavor=key,
                        algo=algo_,
                    )
                    if path is None:
                        continue
                    dataset_name = f"{selection_}_{key}_{algo_}"
                    dataset = cls(
                        name=dataset_name,
                        project=project_name,
                        flavor=key,
                        algo=algo_,
                        tag="test",
                        selection=selection_,
                    )
                    datasets.append(dataset)
                    dataset_list_dict[dataset_key].append(dataset_name)

        for ds_name, ds_list in dataset_list_dict.items():
            # Skip empty lists
            if not ds_list:
                continue
            dataset_list = RailDatasetListHolder(
                name=ds_name,
                dataset_class=cls.output_type.full_class_name(),
                datasets=ds_list,
            )
            dataset_lists.append(dataset_list)

        return (projects, datasets, dataset_lists)


class RailPZMultiPointEstimateDataHolder(RailDatasetHolder):
    """Simple class for holding making a merged for plotting data that comes from a RailProject"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Dataset name"),
        datasets=StageParameter(
            list, None, fmt="%s", required=True, msg="Dataset name"
        ),
    )

    extractor_inputs: dict = {
        "datasets": list[RailDatasetHolder],
    }

    output_type: type[RailDataset] = RailPZMultiPointEstimateDataset

    def __init__(self, **kwargs: Any):
        RailDatasetHolder.__init__(self, **kwargs)
        self._datasets: list[RailDatasetHolder] | None = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}( {self.config.datasets} )"

    def _get_data(self, **kwargs: Any) -> dict[str, Any] | None:
        the_datasets = kwargs.get("datasets", None)
        if the_datasets is None:  # pragma: no cover
            raise KeyError(f"Missed datasets {kwargs}")
        point_estimate_infos: dict[str, dict[str, Any]] = {}
        for dataset_ in the_datasets:
            the_name = dataset_.config.name
            point_estimate_infos[the_name] = dataset_.get_extractor_inputs()
        return get_multi_pz_point_estimate_data(point_estimate_infos)

    def get_extractor_inputs(self) -> dict[str, Any]:
        if self._datasets is None:
            self._datasets = [
                RailDatasetFactory.get_dataset(key) for key in self.config.datasets
            ]

        the_extractor_inputs = dict(
            datasets=self._datasets,
        )
        self._validate_extractor_inputs(**the_extractor_inputs)
        return the_extractor_inputs



from click.testing import CliRunner, Result

from rail.cli.rail_plot.plot_commands import plot_cli


def check_result(
    result: Result,
) -> None:
    if not result.exit_code == 0:
        raise ValueError(f"{result} failed with {result.exit_code} {result.output}")


def test_cli_help() -> None:
    runner = CliRunner()

    result = runner.invoke(plot_cli, "--help")
    check_result(result)


def test_cli_run(setup_project_area: int) -> None:
    assert setup_project_area == 0
    runner = CliRunner()

    result = runner.invoke(
        plot_cli, "run --outdir test/temp_dir/plots/ tests/ci_plot_groups.yaml"
    )
    check_result(result)


def test_cli_inspect(setup_project_area: int) -> None:
    assert setup_project_area == 0
    runner = CliRunner()

    result = runner.invoke(plot_cli, "inspect tests/ci_plot_groups.yaml")
    check_result(result)


def test_cli_extract_nz_datasets(setup_project_area: int) -> None:
    assert setup_project_area == 0
    runner = CliRunner()

    # run with split by algo
    result = runner.invoke(
        plot_cli,
        "extract-datasets "
        "--dataset-holder-class rail.plotting.nz_data_holders.RailNZTomoBinsDataHolder "
        "--flavor all "
        "--split-mode by_algo "
        "--output-yaml tests/temp_data/dataset_nz_out.yaml "
        "tests/ci_project.yaml",
    )
    check_result(result)

    # run with split by flavor
    result = runner.invoke(
        plot_cli,
        "extract-datasets "
        "--dataset-holder-class rail.plotting.nz_data_holders.RailNZTomoBinsDataHolder "
        "--flavor all "
        "--split-mode by_flavor "
        "--output-yaml tests/temp_data/dataset_nz_out.yaml "
        "tests/ci_project.yaml",
    )
    check_result(result)

    # run without split
    result = runner.invoke(
        plot_cli,
        "extract-datasets "
        "--dataset-holder-class rail.plotting.nz_data_holders.RailNZTomoBinsDataHolder "
        "--flavor all "
        "--split-mode no_split "
        "--output-yaml tests/temp_data/dataset_nz_out.yaml "
        "tests/ci_project.yaml",
    )
    check_result(result)


def test_cli_extract_pz_datasets(setup_project_area: int) -> None:
    assert setup_project_area == 0
    runner = CliRunner()

    # run with split by algo
    result = runner.invoke(
        plot_cli,
        "extract-datasets "
        "--dataset-holder-class rail.plotting.pz_data_holders.RailPZPointEstimateDataHolder "
        "--flavor all "
        "--split-mode by_algo "
        "--output-yaml tests/temp_data/dataset_out.yaml "
        "tests/ci_project.yaml",
    )
    check_result(result)

    # run with split by flavor
    result = runner.invoke(
        plot_cli,
        "extract-datasets "
        "--dataset-holder-class rail.plotting.pz_data_holders.RailPZPointEstimateDataHolder "
        "--flavor all "
        "--split-mode by_flavor "
        "--output-yaml tests/temp_data/dataset_out.yaml "
        "tests/ci_project.yaml",
    )
    check_result(result)

    # run without split
    result = runner.invoke(
        plot_cli,
        "extract-datasets "
        "--dataset-holder-class rail.plotting.pz_data_holders.RailPZPointEstimateDataHolder "
        "--flavor all "
        "--split-mode no_split "
        "--output-yaml tests/temp_data/dataset_out.yaml "
        "tests/ci_project.yaml",
    )
    check_result(result)


def test_cli_make_plot_groups(setup_project_area: int) -> None:
    assert setup_project_area == 0
    runner = CliRunner()

    result = runner.invoke(
        plot_cli,
        "make-plot-groups-for-dataset-list "
        "--output-yaml tests/temp_data/check_plot_group.yaml "
        "--plotter-yaml-path tests/ci_plots.yaml "
        "--dataset-yaml-path tests/ci_datasets.yaml "
        "--plotter-list-name zestimate_v_ztrue "
        "--dataset-list-name blend_baseline_all",
    )
    check_result(result)


def test_cli_make_nz_plot_groups(setup_project_area: int) -> None:
    assert setup_project_area == 0
    runner = CliRunner()

    result = runner.invoke(
        plot_cli,
        "make-plot-groups-for-dataset-list "
        "--output-yaml tests/temp_data/check_nz_plot_group.yaml "
        "--plotter-yaml-path tests/ci_plots.yaml "
        "--dataset-yaml-path tests/ci_datasets.yaml "
        "--plotter-list-name tomo_bins "
        "--dataset-list-name blend_baseline_tomo_knn",
    )
    check_result(result)


def test_cli_make_plot_groups_for_project(setup_project_area: int) -> None:
    assert setup_project_area == 0
    runner = CliRunner()

    result = runner.invoke(
        plot_cli,
        "make-plot-groups-for-project "
        "--output-yaml tests/temp_data/check_project_plot_groups_by_algo.yaml "
        "--plotter-yaml-path tests/ci_plots.yaml "
        "--flavor all "
        "--split-mode by_algo "
        "tests/ci_project.yaml",
    )
    check_result(result)

    result = runner.invoke(
        plot_cli,
        "make-plot-groups-for-project "
        "--output-yaml tests/temp_data/check_project_plot_groups_by_flavor.yaml "
        "--plotter-yaml-path tests/ci_plots.yaml "
        "--flavor all "
        "--split-mode by_flavor "
        "tests/ci_project.yaml",
    )
    check_result(result)



from __future__ import annotations

from typing import Any

from ceci.config import StageParameter
from rail.core.configurable import Configurable
from rail.core.factory_mixin import RailFactoryMixin


class RailSelection(Configurable):
    """Paramters for a simple data selection

    This is just defined as a dict of cuts
    """

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Selection name"),
        cuts=StageParameter(
            dict,
            {},
            fmt="%s",
            msg="Cuts associated to selection",
        ),
    )
    yaml_tag = "Selection"

    def __init__(self, **kwargs: Any):
        """C'tor

        Parameters
        ----------
        **kwargs
            Configuration parameters for this RailAlgorithmHolder, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)

    def __repr__(self) -> str:
        return f"cuts={self.config.cuts}"


class RailSelectionFactory(RailFactoryMixin):
    """Factory class to make selections

    Expected usage is that user will define a yaml file with the various
    datasets that they wish to use with the following example syntax:

    .. highlight:: yaml
    .. code-block:: yaml

      Selections:
        - Selection:
            name: maglim_25.5
            cuts:
              maglim_i: [null, 25.5]
    """

    yaml_tag = "Selections"

    client_classes = [RailSelection]

    _instance: RailSelectionFactory | None = None

    def __init__(self) -> None:
        """C'tor, build an empty RailDatasetFactory"""
        RailFactoryMixin.__init__(self)
        self._selections = self.add_dict(RailSelection)

    @classmethod
    def get_selections(cls) -> dict[str, RailSelection]:
        """Return the dict of all the selections"""
        return cls.instance().selections

    @classmethod
    def get_selection_names(cls) -> list[str]:
        """Return the names of the selections"""
        return list(cls.instance().selections.keys())

    @classmethod
    def get_selection(cls, name: str) -> RailSelection:
        """Get a selection by it's assigned name

        Parameters
        ----------
        name: str
            Name of the selection templates to return

        Returns
        -------
        RailSelection
            selection in question
        """
        try:
            return cls.instance().selections[name]
        except KeyError as msg:
            raise KeyError(
                f"RailSelection named {name} not found in RailSelectionFactory "
                f"{list(cls.instance().selections.keys())}"
            ) from msg

    @classmethod
    def add_selection(cls, selection: RailSelection) -> None:
        cls.instance().add_to_dict(selection)

    @property
    def selections(self) -> dict[str, RailSelection]:
        """Return the dictionary of selection templates"""
        return self._selections



from typing import Any

import click
from rail.cli.rail import options
from rail.core import __version__

from rail.cli.rail_project import project_options
from rail.plotting import control

from . import plot_options

__all__ = [
    "plot_cli",
    "run_command",
    "inspect_command",
    "extract_datasets_command",
    "make_plot_groups_for_dataset_list",
    "make_plot_groups_for_project",
]


@click.group()
@click.version_option(__version__)
def plot_cli() -> None:
    """RAIL plotting functions

    These commands to work with RailProject infrastructure to generate
    sets of standard plots and html pages to help browse them.

    The configuration file can include these yaml_tags

    1. `Plots` with type of plots available
    2. `Data` with specific datasets we can make those plots with
    3. `PlotGroup` with combinations of the two
    """


@plot_cli.command(name="run")
@project_options.config_file()
@plot_options.include_groups()
@plot_options.exclude_groups()
@plot_options.save_plots()
@plot_options.purge_plots()
@plot_options.find_only()
@plot_options.make_html()
@options.outdir()
def run_command(config_file: str, **kwargs: Any) -> int:
    """Make a bunch of plots

    The configuration file should define both the plots to make
    and the datasets to use.
    """
    control.clear()
    control.run(config_file, **kwargs)
    return 0


@plot_cli.command(name="inspect")
@project_options.config_file()
def inspect_command(config_file: str) -> int:
    """Inspect a configuration yaml file

    These will load the configuration file, and any files that that it includes
    and then print out the contents of the component library.
    """
    control.clear()
    control.load_yaml(config_file)
    control.print_contents()
    return 0


@plot_cli.command(name="extract-datasets")
@project_options.config_file()
@plot_options.dataset_holder_class()
@plot_options.output_yaml()
@plot_options.dataset_list_name()
@project_options.flavor()
@project_options.selection()
@plot_options.split_mode()
def extract_datasets_command(
    config_file: str,
    dataset_holder_class: str,
    output_yaml: str,
    **kwargs: dict[str, Any],
) -> int:
    """Create a yaml file with the datasets in a project

    This will read a yaml project configuration file and
    search the project for all the datasets that the
    extractor_class is able to extract and write the
    results to the output_yaml file.
    """

    control.clear()
    control.extract_datasets(
        config_file,
        dataset_holder_class,
        output_yaml=output_yaml,
        **kwargs,
    )
    return 0


@plot_cli.command(name="make-plot-groups-for-dataset-list")
@plot_options.output_yaml()
@plot_options.plotter_yaml_path()
@plot_options.dataset_yaml_path()
@plot_options.plotter_list_name()
@plot_options.output_prefix()
@plot_options.dataset_list_name(multiple=True)
def make_plot_groups_for_dataset_list(
    output_yaml: str, **kwargs: dict[str, Any]
) -> int:
    """Combine plotters with availble datsets

    This will read the plotter_yaml and dataset_yaml
    files, and combine all the datasets in the list
    given by dataset_list_name with all the plots given
    by plotter_list_name and write the results to the output_yaml
    file.
    """
    control.clear()
    control.make_plot_group_yaml_for_dataset_list(output_yaml, **kwargs)
    return 0


@plot_cli.command(name="make-plot-groups-for-project")
@plot_options.output_yaml()
@project_options.config_file()
@plot_options.plotter_yaml_path()
@project_options.flavor()
@project_options.selection()
@plot_options.split_mode()
def make_plot_groups_for_project(
    output_yaml: str,
    plotter_yaml_path: str,
    config_file: str,
    **kwargs: dict[str, Any],
) -> int:
    """Combine plotters with availble datsets

    This will read the plotter_yaml and dataset_yaml
    files, and combine all the datasets in the list
    given by dataset_list_name with all the plots given
    by plotter_list_name and write the results to the output_yaml
    file.
    """
    control.clear()
    control.make_plot_group_yaml_for_project(
        output_yaml, plotter_yaml_path, config_file, **kwargs
    )
    return 0



from __future__ import annotations

from rail.core.factory_mixin import RailFactoryMixin

from .file_template import RailProjectFileInstance, RailProjectFileTemplate


class RailProjectFileFactory(RailFactoryMixin):
    """Factory class to make files

    Expected usage is that user will define a yaml file with the various
    datasets that they wish to use with the following example syntax:

    .. highlight:: yaml
    .. code-block:: yaml

      Files:
        - FileTemplate:
            name: test_file_100k
            path_template: "{catalogs_dir}/test/{project}_{selection}_baseline_100k.hdf5"

    Or the used can specifiy particular file instances where everything except the
    interation_vars are resolved

    .. highlight:: yaml
    .. code-block:: yaml

      Files:
        - FileInstance
            name: test_file_100k_roman_rubin_v1.1.3_gold
            path: <full_path_to_file>
    """

    yaml_tag: str = "Files"

    client_classes = [RailProjectFileInstance, RailProjectFileTemplate]

    _instance: RailProjectFileFactory | None = None

    def __init__(self) -> None:
        """C'tor, build an empty RailProjectFileFactory"""
        RailFactoryMixin.__init__(self)
        self._file_templates = self.add_dict(RailProjectFileTemplate)
        self._file_instances = self.add_dict(RailProjectFileInstance)

    @classmethod
    def get_file_templates(cls) -> dict[str, RailProjectFileTemplate]:
        """Return the dict of all the file templates"""
        return cls.instance().file_templates

    @classmethod
    def get_file_template_names(cls) -> list[str]:
        """Return the names of the file templates"""
        return list(cls.instance().file_templates.keys())

    @classmethod
    def get_file_instances(cls) -> dict[str, RailProjectFileInstance]:
        """Return the dict of all the file instances"""
        return cls.instance().file_instances

    @classmethod
    def get_file_instance_names(cls) -> list[str]:
        """Return the names of the file instances lists"""
        return list(cls.instance().file_instances.keys())

    @classmethod
    def get_file_template(cls, name: str) -> RailProjectFileTemplate:
        """Get file template by it's assigned name

        Parameters
        ----------
        name: str
            Name of the file templates to return

        Returns
        -------
        RailProjectFileTemplate:
            file templates in question
        """
        try:
            return cls.instance().file_templates[name]
        except KeyError as msg:
            raise KeyError(
                f"Dataset named {name} not found in RailProjectFileFactory "
                f"{list(cls.instance().file_templates.keys())}"
            ) from msg

    @classmethod
    def get_file_instance(cls, name: str) -> RailProjectFileInstance:
        """Get a file instance by its assigned name

        Parameters
        ----------
        name: str
            Name of the file instance list to return

        Returns
        -------
        RailProjectFileInstance:
            file instance in question
        """
        try:
            return cls.instance().file_instances[name]
        except KeyError as msg:
            raise KeyError(
                f"RailProjectFileInstance named {name} not found in RailProjectFileInstance "
                f"{list(cls.instance().file_instances.keys())}"
            ) from msg

    @classmethod
    def add_file_instance(cls, file_instance: RailProjectFileInstance) -> None:
        """Add a particular RailProjectFileInstance to the factory"""
        cls.instance().add_to_dict(file_instance)

    @classmethod
    def add_file_template(cls, file_template: RailProjectFileTemplate) -> None:
        """Add a particular RailProjectFileTemplate to the factory"""
        cls.instance().add_to_dict(file_template)

    @property
    def file_templates(self) -> dict[str, RailProjectFileTemplate]:
        """Return the dictionary of file templates"""
        return self._file_templates

    @property
    def file_instances(self) -> dict[str, RailProjectFileInstance]:
        """Return the dictionary of file instances"""
        return self._file_instances

    def print_instance_contents(self) -> None:
        """Print the contents of the factory"""
        print("----------------")
        print("Files:")
        RailFactoryMixin.print_instance_contents(self)



"""Class to get track of groups of plots to make"""

from __future__ import annotations

import os
from typing import Any

from ceci.config import StageParameter
from jinja2 import Environment, FileSystemLoader, Template
from rail.core.configurable import Configurable

from .dataset import RailDataset
from .dataset_factory import RailDatasetFactory
from .dataset_holder import RailDatasetHolder
from .plot_holder import RailPlotDict, RailPlotHolder
from .plotter import RailPlotter
from .plotter_factory import RailPlotterFactory

HTML_TEMPLATE_DIR = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "html_templates"
)


class RailPlotGroup(Configurable):
    """Class defining of a group on plots to make
    with a particular list of coherent datasets

    """

    jinja_env: Environment | None = None
    jinja_template: Template | None = None
    jinja_index_template: Template | None = None

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(
            str, None, fmt="%s", required=True, msg="PlotGroupName name"
        ),
        plotter_list_name=StageParameter(
            str, None, fmt="%s", required=True, msg="PlotterList name"
        ),
        dataset_list_name=StageParameter(
            str, None, fmt="%s", required=True, msg="DatasetList name"
        ),
        outdir=StageParameter(str, ".", fmt="%s", msg="Output directory"),
        figtype=StageParameter(str, "png", fmt="%s", msg="Plot type"),
    )

    yaml_tag = "PlotGroup"

    @classmethod
    def _load_jinja(cls) -> None:
        if cls.jinja_template is not None:  # pragma: no cover
            return
        cls.jinja_env = Environment(loader=FileSystemLoader(HTML_TEMPLATE_DIR))
        cls.jinja_template = cls.jinja_env.get_template("plot_group_table.html")
        cls.jinja_index_template = cls.jinja_env.get_template("plot_group_index.html")

    def __init__(self, **kwargs: Any) -> None:
        Configurable.__init__(self, **kwargs)
        self._plots: dict[str, RailPlotDict] = {}
        self._plotter_list: list[RailPlotter] = []
        self._dataset_list: list[RailDatasetHolder] = []

    def __repr__(self) -> str:
        return f"PlotGroup: {self.config.plotter_list_name}, DatasetList: {self.config.dataset_list_name}"

    @property
    def plotter_list(self) -> list[RailPlotter]:
        return self._plotter_list

    @property
    def dataset_list(self) -> list[RailDatasetHolder]:
        return self._dataset_list

    def find_plot(self, dataset_name: str, plotter_name: str) -> RailPlotHolder:
        """Find a particular plot"""
        try:
            sub_dict = self._plots[dataset_name]
        except KeyError as msg:
            raise KeyError(
                f"Failed to find {dataset_name} in {list(self._plots.keys())}"
            ) from msg
        assert sub_dict.plots
        try:
            return sub_dict.plots[plotter_name]
        except KeyError as msg:
            raise KeyError(
                f"Failed to find {plotter_name} in {list(sub_dict.plots.keys())}"
            ) from msg

    def find_plot_path(self, dataset_name: str, plotter_name: str) -> str | None:
        """Find a particular plot and get the path to the associated file"""
        return self.find_plot(dataset_name, plotter_name).path

    def resolve(self) -> None:
        """Get the PlotterList and DatasetList"""
        plotter_factory = RailPlotterFactory.instance()
        dataset_factory = RailDatasetFactory.instance()
        plotter_list_wrapper = plotter_factory.get_plotter_list(
            self.config.plotter_list_name
        )
        dataset_list_wrapper = dataset_factory.get_dataset_list(
            self.config.dataset_list_name
        )
        plotter_data_holder_class = RailDatasetHolder.load_sub_class(
            plotter_list_wrapper.config.dataset_holder_class
        )
        plotter_input_class = plotter_data_holder_class.output_type
        dataset_output_class = RailDataset.load_sub_class(
            dataset_list_wrapper.config.dataset_class
        )
        if not issubclass(
            dataset_output_class, plotter_input_class
        ):  # pragma: no cover
            raise TypeError(
                f"RailPlotGroup {self.config.name} dataset output class {dataset_output_class} "
                f"is not a subclass of plotter input class {plotter_input_class}"
            )
        self._plotter_list = plotter_list_wrapper.resolve(plotter_factory)
        self._dataset_list = dataset_list_wrapper.resolve(dataset_factory)

    def make_plots(
        self,
    ) -> dict[str, RailPlotDict]:
        """Make a set of plots

        Returns
        -------
        out_dict: dict[str, RailPlotDict]
            Dictionary of the newly created figures
        """
        self.resolve()
        self._plots.update(
            **RailPlotter.iterate(self._plotter_list, self._dataset_list)
        )
        return self._plots

    def find_plots(
        self,
        outdir: str,
    ) -> dict[str, RailPlotDict]:
        """Find a set of plots

        Parameters
        ----------
        outdir: str
            Prepend this to the groups output dir

        Returns
        -------
        out_dict: dict[str, Figure]
            Dictionary of the newly created figures
        """
        self.resolve()
        self._plots.update(
            **RailPlotter.iterate(
                self._plotter_list,
                self._dataset_list,
                find_only=True,
                outdir=outdir,
                figtype=self.config.figtype,
            )
        )
        return self._plots

    @classmethod
    def make_html_index(
        cls,
        outfile: str,
        output_pages: list[str],
    ) -> None:
        """Make the html index page for a list of other pages

        Parameters
        ----------
        outfile: str
            Html file to write

        output_pages: list[str]
            Set of pages to include in the index
        """
        cls._load_jinja()
        assert cls.jinja_index_template is not None

        # Render template  data and save to HTML file
        output = cls.jinja_index_template.render(output_pages=output_pages, os=os)
        with open(outfile, "w", encoding="utf-8") as file:
            file.write(output)

    def make_html(
        self,
        outfile: str,
    ) -> None:
        """Make the html page to display the plots made by this plot group

        Parameters
        ----------
        outfile: str
            Html file to write
        """

        self._load_jinja()
        assert self.jinja_template is not None

        # Render template  data and save to HTML file
        output = self.jinja_template.render(plot_group=self, os=os)
        with open(outfile, "w", encoding="utf-8") as file:
            file.write(output)

    def run(
        self,
        save_plots: bool = True,
        purge_plots: bool = True,
        find_only: bool = False,
        outdir: str | None = None,
        make_html: bool = False,
        output_html: str | None = None,
    ) -> dict[str, RailPlotDict]:
        """Make all the plots given the data

        Parameters
        ----------
        save_plots: bool
            If true, save the plots to disk

        purge_plots: bool
            If true, delete the plots after saving

        find_only: bool
            If true, only look for existing plots

        make_html: bool
            If true, make an html table to browse plots

        outdir: str | None
            If set, prepend this to the groups output dir

        output_html: str | None
            Path for output html file

        Returns
        -------
        out_dict: dict[str, Figure]
            Dictionary of the newly created figures
        """
        if outdir is not None:
            output_dir = os.path.join(outdir, self.config.outdir)
        else:  # pragma: no cover
            output_dir = self.config.outdir

        if find_only:
            self.find_plots(
                outdir=output_dir,
            )
        else:
            self.make_plots()
            if save_plots:
                RailPlotter.write_plots(
                    self._plots, output_dir, self.config.figtype, purge=purge_plots
                )
        if make_html:
            if output_html is None:
                assert outdir
                output_html = os.path.join(outdir, f"plots_{self.config.name}.html")
            self.make_html(output_html)
        return self._plots



from __future__ import annotations

from rail.core.factory_mixin import RailFactoryMixin

from .catalog_template import RailProjectCatalogInstance, RailProjectCatalogTemplate


class RailCatalogFactory(RailFactoryMixin):
    """Factory class to make catalogs

    Expected usage is that user will define a yaml file with the various
    datasets that they wish to use with the following example syntax:

    .. highlight:: yaml
    .. code-block:: yaml

      Catalogs:
        - CatalogTemplate
            name: truth
            path_template: "{catalogs_dir}/{project}_{sim_version}/{healpix}/part-0.parquet"
            iteration_vars: ['healpix']
        - CatalogTemplate
            name: reduced
            path_template: "{catalogs_dir}/{project}_{sim_version}_{selection}/{healpix}/part-0.pq"
            iteration_vars: ['healpix']

    Or the used can specifiy particular catalog instances where everything except the
    interation_vars are resolved

    .. highlight:: yaml
    .. code-block:: yaml

      Catalogs:
        - CatalogTemplate
            name: truth_roman_rubin_v1.1.3_gold
            path_template: "full_path_to_catalog/{healpix}/part-0.parquet"
            iteration_vars: ['healpix']
    """

    yaml_tag: str = "Catalogs"

    client_classes = [RailProjectCatalogTemplate, RailProjectCatalogInstance]

    _instance: RailCatalogFactory | None = None

    def __init__(self) -> None:
        """C'tor, build an empty RailDatasetFactory"""
        RailFactoryMixin.__init__(self)
        self._catalog_templates = self.add_dict(RailProjectCatalogTemplate)
        self._catalog_instances = self.add_dict(RailProjectCatalogInstance)

    @classmethod
    def get_catalog_templates(cls) -> dict[str, RailProjectCatalogTemplate]:
        """Return the dict of all the catalog templates"""
        return cls.instance().catalog_templates

    @classmethod
    def get_catalog_template_names(cls) -> list[str]:
        """Return the names of the catalog templates"""
        return list(cls.instance().catalog_templates.keys())

    @classmethod
    def get_catalog_instances(cls) -> dict[str, RailProjectCatalogInstance]:
        """Return the dict of all the catalog instances"""
        return cls.instance().catalog_instances

    @classmethod
    def get_catalog_instance_names(cls) -> list[str]:
        """Return the names of the catalog instances lists"""
        return list(cls.instance().catalog_instances.keys())

    @classmethod
    def get_catalog_template(cls, name: str) -> RailProjectCatalogTemplate:
        """Get catalog templates by it's assigned name"""
        try:
            return cls.instance().catalog_templates[name]
        except KeyError as msg:
            raise KeyError(
                f"Dataset named {name} not found in RailCatalogFactory "
                f"{list(cls.instance().catalog_templates.keys())}"
            ) from msg

    @classmethod
    def get_catalog_instance(cls, name: str) -> RailProjectCatalogInstance:
        """Get a catalog instance by its assigned name"""
        try:
            return cls.instance().catalog_instances[name]
        except KeyError as msg:
            raise KeyError(
                f"RailProjectCatalogInstance named {name} not found in RailProjectCatalogInstance "
                f"{list(cls.instance().catalog_instances.keys())}"
            ) from msg

    @classmethod
    def add_catalog_instance(cls, catalog_instance: RailProjectCatalogInstance) -> None:
        """Add a particular catalog instance to the factory"""
        cls.instance().add_to_dict(catalog_instance)

    @classmethod
    def add_catalog_template(cls, catalog_template: RailProjectCatalogTemplate) -> None:
        """Add a particular catalog template to the factory"""
        cls.instance().add_to_dict(catalog_template)

    @property
    def catalog_templates(self) -> dict[str, RailProjectCatalogTemplate]:
        """Return the dictionary of catalog templates"""
        return self._catalog_templates

    @property
    def catalog_instances(self) -> dict[str, RailProjectCatalogInstance]:
        """Return the dictionary of catalog instances"""
        return self._catalog_instances

    def print_instance_contents(self) -> None:
        """Print the contents of the factory"""
        print("----------------")
        print("Catalogs")
        RailFactoryMixin.print_instance_contents(self)
